@@ -3,10 +3,11 @@ package IO::Socket::Socks::Wrapper;
 use strict;
 no warnings 'prototype';
 no warnings 'redefine';
-use Socket;
+use Socket qw(:DEFAULT inet_ntop);
+use Errno;
 use base 'Exporter';
 
-our $VERSION = '0.12';
+our $VERSION = '0.13';
 our @EXPORT_OK = ('connect', 'wrap_connection');
 
 # cache
@@ -21,51 +22,51 @@ sub _io_socket_connect_ref {
 }
 
 # fake handle to put under event loop while making socks handshake
-my ($blocking_reader, $blocking_writer);
-sub _get_blocking_handle {
-	unless ($blocking_reader) {
-		pipe($blocking_reader, $blocking_writer)
-			or die 'pipe(): ', $!;
+sub _get_blocking_handles {
+	pipe(my $blocking_reader, my $blocking_writer)
+		or die 'socketpair(): ', $!;
+	
+	$blocking_writer->blocking(0);
+	$blocking_reader->blocking(0);
+	my $garbage = '\0' x 1024;
+	
+	my ($writed, $total_writed);
+	while ($writed = syswrite($blocking_writer, $garbage)) {
+		$total_writed += $writed;
 		
-		vec(my $win, fileno($blocking_reader), 1) = 1;
-		if (select(undef, $win, undef, 0)) {
-			# oh shi~, this system is crazy
-			# reader from pipe() marked as ready for write
-			($blocking_reader, $blocking_writer) = ($blocking_writer, $blocking_reader);
-			$blocking_reader->blocking(0);
-			
-			my $garbage = '\0' x 4096;
-			my ($writed, $total_writed);
-			while ($writed = syswrite($blocking_reader, $garbage)) {
-				$total_writed += $writed;
-				
-				if ($total_writed > 2097152) {
-					# pipe with buffer more than 2 mb
-					# are u kidding me?
-					die "Can't create blocking handle";
-				}
-			}
+		if ($total_writed > 2097152) {
+			# socket with buffer more than 2 mb
+			# are u kidding me?
+			die "Can't create blocking handle";
 		}
 	}
 	
-	return $blocking_reader;
+	return ($blocking_reader, $blocking_writer);
 }
 
-sub import
-{
+sub _unblock_handles {
+	my ($blocking_reader, $blocking_writer) = @_;
+	
+	while (sysread($blocking_reader, my $buf, 4096)) {
+		vec(my $win, fileno($blocking_writer), 1) = 1;
+		last if select(undef, $win, undef, 0);
+	}
+}
+
+sub import {
 	my $mypkg = shift;
 	
 	if (@_ == 1 && !ref($_[0]) && $_[0] eq 'wrap_connection') {
 		return __PACKAGE__->export_to_level(1, $mypkg, 'wrap_connection');
 	}
 	
-	while(my ($pkg, $cfg) = splice @_, 0, 2) {
-		unless(defined $cfg) {
+	while (my ($pkg, $cfg) = splice @_, 0, 2) {
+		unless (defined $cfg) {
 			$cfg = $pkg;
 			$pkg = undef;
 		}
 		
-		if($pkg) {
+		if ($pkg) {
 			no strict 'refs';
 			
 			my $sub;
@@ -94,7 +95,7 @@ sub import
 				_io_socket_connect_ref();
 				
 				*$symbol = sub {
-					local *IO::Socket::connect = sub {
+					local *IO::Socket::IP::connect = local *IO::Socket::connect = sub {
 						_connect(@_, $cfg, 1);
 					};
 					
@@ -102,7 +103,7 @@ sub import
 				};
 				next;
 			}
-			elsif($pkg->isa('IO::Socket')) {
+			elsif ($pkg->isa('IO::Socket')) {
 			# replace IO::Socket::connect
 			# if package inherits from IO::Socket
 				# save replaceable package version of the connect
@@ -116,7 +117,7 @@ sub import
 				*connect = sub {
 					_io_socket_connect_ref();
 					
-					local(*IO::Socket::connect) = sub {
+					local *IO::Socket::IP::connect = local *IO::Socket::connect = sub {
 						_connect(@_, $cfg, 1);
 					};
 					
@@ -176,12 +177,11 @@ sub wrap_connection {
 	return  IO::Socket::Socks::Wrapped->new(@_);
 }
 
-sub _connect
-{
+sub _connect {
 	my ($socket, $name, $cfg, $io_socket) = @_;
 	my $ref = ref($socket);
 	
-	if (($ref && $socket->isa('IO::Socket::Socks')) || !$cfg) {
+	if ($socket->isa('IO::Socket::Socks') || !$cfg) {
 		unless ($io_socket and ${*$socket}{'io_socket_timeout'}) {
 			return CORE::connect( $socket, $name );
 		}
@@ -191,8 +191,14 @@ sub _connect
 		return _io_socket_connect_ref->( $socket, $name );
 	}
 	
-	my ($port, $host) = sockaddr_in($name);
-	$host = inet_ntoa($host);
+	my ($port, $host);
+	if (($port, $host) = eval { unpack_sockaddr_in($name) }) {
+		$host = inet_ntoa($host);
+	}
+	else {
+		($port, $host) = unpack_sockaddr_in6($name);
+		$host = inet_ntop(AF_INET6, $host);
+	}
 	
 	# global overriding will not work with `use'
 	require IO::Socket::Socks;
@@ -202,51 +208,77 @@ sub _connect
 		$cfg->{Timeout} = $ref && $socket->isa('IO::Socket') && ${*$socket}{'io_socket_timeout'} || 180;
 	}
 	
+	my $need_nb;
+	
 	if ($io_handler) {
 		$io_handler = $io_handler->();
 		require POSIX;
-		require Scalar::Util;
+		
 		
 		my $fd = fileno($socket);
-		my $tmp_fd = POSIX::dup($fd) // die 'dup2(): ', $!;
+		my $tmp_fd = POSIX::dup($fd) // die 'dup(): ', $!;
 		open my $tmp_socket, '+<&=' . $tmp_fd or die 'open(): ', $!;
-		POSIX::dup2(fileno(_get_blocking_handle()), $fd) // die 'dup(): ', $!;
+		
+		my ($blocking_reader, $blocking_writer) = _get_blocking_handles();
+		POSIX::dup2(fileno($blocking_writer), $fd) // die 'dup2(): ', $!;
+		
+		$io_handler->{blocking_reader} = $blocking_reader;
+		$io_handler->{blocking_writer} = $blocking_writer;
 		$io_handler->{orig_socket} = $socket;
 		Scalar::Util::weaken($io_handler->{orig_socket});
 		$socket = $tmp_socket;
 	}
+	elsif (!$socket->blocking) {
+		# without io handler non-blocking connection will not be success
+		# so set socket to blocking mode while making socks handshake
+		$socket->blocking(1);
+		$need_nb = 1;
+	}
 	
-	IO::Socket::Socks->new_from_socket(
+	my $ok = IO::Socket::Socks->new_from_socket(
 		$socket,
 		ConnectAddr  => $host,
 		ConnectPort  => $port,
 		%$cfg
-	) or return;
+	);
 	
+	if ($need_nb) {
+		$socket->blocking(0);
+	}
+	
+	return unless $ok;
 	bless $socket, $ref if $ref && !$io_handler; # XXX: should we unbless for GLOB?
 	
 	if ($io_handler) {
 		my ($r_cb, $w_cb); 
+		my $done;
 		
 		tie *{$io_handler->{orig_socket}}, 'IO::Socket::Socks::Wrapper::Handle', $io_handler->{orig_socket}, sub {
-			$io_handler->{unset_read_watcher}->($socket);
-			$io_handler->{unset_write_watcher}->($socket);
-			
-			if ($io_handler->{destroy_io_watcher}) {
-				$io_handler->{destroy_io_watcher}->($socket);
+			unless ($done) {
+				$io_handler->{unset_read_watcher}->($socket);
+				$io_handler->{unset_write_watcher}->($socket);
+				
+				if ($io_handler->{destroy_io_watcher}) {
+					$io_handler->{destroy_io_watcher}->($socket);
+				}
+				
+				close $socket;
 			}
 			
-			close $socket;
+			# clean circular references
+			undef $r_cb;
+			undef $w_cb;
 		};
 		
 		my $on_finish = sub {
-			tied(*{$io_handler->{orig_socket}})->handshake_done(1);
+			tied(*{$io_handler->{orig_socket}})->handshake_done($done = 1);
 			POSIX::dup2(fileno($socket), fileno($io_handler->{orig_socket})) // die 'dup2(): ', $!;
 			close $socket;
+			_unblock_handles($io_handler->{blocking_reader}, $io_handler->{blocking_writer});
 		};
 		
 		my $on_error = sub {
-			tied(*{$io_handler->{orig_socket}})->handshake_done(1);
+			tied(*{$io_handler->{orig_socket}})->handshake_done($done = 1);
 			shutdown($socket, 0);
 			POSIX::dup2(fileno($socket), fileno($io_handler->{orig_socket}))  // die 'dup2(): ', $!;
 			close $socket;
@@ -307,6 +339,9 @@ sub _connect
 		}
 		
 		$io_handler->{set_write_watcher}->($socket, $w_cb);
+		
+		$! = Errno::EINPROGRESS;
+		return 0;
 	}
 	
 	return 1;
@@ -366,11 +401,7 @@ sub CLOSE {
 
 sub DESTROY {
 	my $self = shift;
-	
-	unless ($self->handshake_done) {
-		$self->handshake_done(1);
-		${*$self}{cleanup_cb}->();
-	}
+	${*$self}{cleanup_cb}->();
 }
 
 1;
@@ -609,6 +640,8 @@ description see above.
 
 =head4 Integration with event loops
 
+B<Note:> integration with C<kqueue> based event loops known to be broken
+
 When you are using some event loop like AnyEvent or POE it is important to prevent any long blocking operations.
 By default IO::Socket::Socks::Wrapper blocks your program while it making connection and SOCKS handshake with a proxy.
 If you are using fast proxy on localhost this is not big problem, because connection to proxy and making SOCKS
@@ -833,9 +866,7 @@ don't want to limit connection attempt time.
 
 =head1 BUGS
 
-Wrapping doesn't work with impure perl packages. WWW::Curl for example.
-
-No support for IPv6 for now. IO::Socket::IP may break wrapping.
+Wrapping doesn't work with XS based modules, where connection done inside C part. WWW::Curl for example.
 
 =head1 SEE ALSO
 
